// Copyright 2023 Universidad Politécnica de Madrid
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//    * Redistributions of source code must retain the above copyright
//      notice, this list of conditions and the following disclaimer.
//
//    * Redistributions in binary form must reproduce the above copyright
//      notice, this list of conditions and the following disclaimer in the
//      documentation and/or other materials provided with the distribution.
//
//    * Neither the name of the Universidad Politécnica de Madrid nor the names of its
//      contributors may be used to endorse or promote products derived from
//      this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

/***********************************************************************************
 *  \file       platform_state_machine_test.cpp
 *  \brief      This file contains the definitions of the tests for the platform state machine.
 *  \authors    Miguel Fernández Cortizas
 *              Pedro Arias Pérez
 *              David Pérez Saura
 *              Rafael Pérez Seguí
 ********************************************************************************/

#include "platform_state_machine.hpp"

#include <iostream>

// #include "core_libs.hpp"
#include "gtest/gtest.h"

TEST(PlatformStateMachineTest, Normal_functionality) {
  as2::Node test_node("test_node");
  as2::PlatformStateMachine state_machine(&test_node);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::DISARMED);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::ARM);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::LANDED);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::TAKE_OFF);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::TAKING_OFF);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::TOOK_OFF);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::FLYING);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::LAND);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::LANDING);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::LANDED);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::LANDED);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::DISARM);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::DISARMED);
}

TEST(PlatformStateMachineTest, EmergencyTest) {
  as2::Node test_node("test_node");
  for (int i = -1; i < static_cast<int>(as2_msgs::msg::PlatformStatus::LANDING); i++) {
    as2::PlatformStateMachine state_machine(&test_node);
    state_machine.setState(i);
    state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::EMERGENCY);
    EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::EMERGENCY);
  }
}

TEST(PlatformStateMachineTest, CorrectFSMActivation) {
  as2::Node test_node("test_node");
  as2::PlatformStateMachine state_machine(&test_node);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::DISARMED);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::ARM);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::LANDED);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::ARM);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::LANDED);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::TAKE_OFF);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::TAKING_OFF);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::LAND);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::TAKING_OFF);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::EMERGENCY);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::EMERGENCY);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::TOOK_OFF);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::EMERGENCY);
}

TEST(PlatformStateMachineTest, TakeoffTest) {
  as2::Node test_node("test_node");
  as2::PlatformStateMachine state_machine(&test_node);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::DISARMED);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::ARM);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::LANDED);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::ARM);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::LANDED);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::TAKE_OFF);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::TAKING_OFF);
  state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::TOOK_OFF);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::FLYING);
  bool response = state_machine.processEvent(as2_msgs::msg::PlatformStateMachineEvent::TAKE_OFF);
  EXPECT_EQ(state_machine.getState().state, as2_msgs::msg::PlatformStatus::FLYING);
  EXPECT_EQ(response, false);
}

int main(int argc, char * argv[])
{
  rclcpp::init(argc, argv);
  ::testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
